/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ThumbsDown.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


const NS_XUL = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul"
const NS_HTML = "http://www.w3.org/1999/xhtml"

var stringBundle;


function $(id)
{
    return document.getElementById(id);
}

function _(s, arg)
{
    // re-assign on first call
    if (stringBundle instanceof Ci.nsIStringBundle)
        _ = getString_nsIStringBundle;
    else
        _ = getString_stringBundle;
    return _.apply(null, arguments);
}

function getString_stringBundle(s, arg)
{
    if (arg === undefined)
        return stringBundle.getString(s);
    else
    {
        arg = Array.prototype.slice.call(arguments, 1);
        return stringBundle.getFormattedString(s, arg);
    }
}

// use chrome://global/content/strres.js to initialize stringBundle
function getString_nsIStringBundle(s, arg)
{
    if (arg === undefined)
        return stringBundle.GetStringFromName(s);
    else
    {
        arg = Array.prototype.slice.call(arguments, 1);
        return stringBundle.formatStringFromName(s, arg, arg.length);
    }
}

// https://developer.mozilla.org/en/Core_JavaScript_1.5_Guide/Inheritance
function extend(child, supertype)
{
   child.prototype.__proto__ = supertype.prototype;
}

function arrayWithValue(length, value)
{
    var arr = [];
    while (length-- > 0)
        arr.push(value);
    return arr;
}

function arrayEquals(a, b, func)
{
    if (a.length !== b.length)
        return false;
    
    var n = a.length;
    if (func)
    {
        for (var i = 0; i < n; i++)
            if (! func(a[i], b[i]))
                return false;
    }
    else
    {
        for (var i = 0; i < n; i++)
            if (a[i] !== b[i])
                return false;
    }
    
    return true;
}

function count(arr, callback)
{
    if (callback === undefined)
        callback = function(b) { return (b); }
    var count = 0;
    for each (var a in arr)
        if (callback(a))
            count++;
    return count;
}

function trim(s)
{
    return s.replace(/^\s+|\s+$/g, "");
}

function replace(template, replacements)
{
    var replacer = function(substring)
    {
        var name = substring.substr(2, substring.length - 3);
        if (name in replacements)
            return replacements[name];
        return substring;
    };
    return template.replace(/\$\{\w+\}/g, replacer);
}

function escapeRegExp(s)
{
    return s.replace(/([.*+?^${}()|[\]\\])/g, '\\$1');
}

function createElement(namespace, name, attrs, children)
{
    var element = document.createElementNS(namespace, name);
    if (attrs)
        for (let [attrName, attrValue] in Iterator(attrs))
            element.setAttribute(attrName, attrValue);
    if (children)
        for each (var child in children)
            element.appendChild(child);
    return element;
}

function createXulElement(name, attrs, children)
{
    return createElement(NS_XUL, name, attrs, children);
}

function createHtmlElement(name, attrs, children)
{
    return createElement(NS_HTML, name, attrs, children);
}

function getTreeSelectionIterator(selection)
{
    var rangeCount = selection.getRangeCount();
    for (var i = 0; i < rangeCount; i++)
    {
        var start = {};
        var end = {};
        selection.getRangeAt(i, start, end);
        for (var k = start.value; k <= end.value; k++)
            yield k;
    }
}

function dfs(node, preOrder)
{
    if (preOrder)
        yield node;
    var child = node.firstChild;
    while (child)
    {
        for each (var descendant in dfs(child, preOrder))
            yield descendant;
        child = child.nextSibling;
    }
    if (! preOrder)
        yield node;
}

function loadContent(url)
{
    var ioSvc = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
    var stream = Cc["@mozilla.org/scriptableinputstream;1"].getService(Ci.nsIScriptableInputStream);
    var channel = ioSvc.newChannel(url, null, null);
    var input = channel.open();
    stream.init(input);
    var content = stream.read(input.available());
    stream.close();
    input.close();
    return content;
}

function getHomePage()
{
    var extId = "thumbsdown@mozdev.org";
    var extMgr = Cc["@mozilla.org/extensions/manager;1"].getService(Ci.nsIExtensionManager);
    var installRdfFile = extMgr.getInstallLocation(extId).getItemFile(extId, "install.rdf");
    var ioSvc = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
    var installRdfUrl = ioSvc.newFileURI(installRdfFile);
    var rdfSvc = Cc["@mozilla.org/rdf/rdf-service;1"].getService(Ci.nsIRDFService);
    var dataSource = rdfSvc.GetDataSourceBlocking(installRdfUrl.spec);
    var subject = rdfSvc.GetResource("urn:mozilla:install-manifest");
    var prefix = "http://www.mozilla.org/2004/em-rdf#";
    var predicate = rdfSvc.GetResource(prefix + "homepageURL");
    var homePage = dataSource.GetTarget(subject, predicate, true).QueryInterface(Ci.nsIRDFLiteral).Value;
    return homePage;
}

function openUrl(url)
{
    var windowMediator = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
    var win = windowMediator.getMostRecentWindow("navigator:browser");
    if (! win)
        return;
    var browser = win.getBrowser();
    browser.selectedTab = browser.addTab(url);
}

function openPrefWindow(pane)
{
    var windowUrl = "chrome://thumbsdown/content/options.xul";
    var windowName = "thumbsdown_prefwindow";
    var windowMediator = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
    var win = windowMediator.getMostRecentWindow(windowName);
    if (win)
    {
        win.showPane(pane);
        win.focus();
    }
    else
    {
        var instantApply = nsPreferences.getBoolPref("browser.preferences.instantApply", false);
        var features = "chrome,titlebar,toolbar,centerscreen," + (instantApply ? "dialog=no" : "dialog=yes");
        window.openDialog(windowUrl, windowName, features, {pane: pane});
    }
}

function openScriptRunnerWindow(focus)
{
    var scriptRunner = Cc["@thumbsdown.mozdev.org/script-runner;1"].getService(Ci.nsISupports).wrappedJSObject;
    scriptRunner.openWindow(focus);
}

function openScriptWizard()
{
    var windowUrl = "chrome://thumbsdown/content/scriptwizard/scriptwizard.xul";
    var windowType = "thumbsdown_scriptWizard";
    var windowMediator = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
    
    var win = windowMediator.getMostRecentWindow(windowType);
    if (win)
    {
        win.focus();
    }
    else
    {
        // FIXME: linux: persist screenX/Y doesn't seem to work if window isn't 'dependent' (but that's not nice in windows)
        // resizable - windows default is non-resizable
        var features = "resizable";
        var args = content.document;
        window.openDialog(windowUrl, null, features, args);
    }
}

function schedule(obj, func, args, delay, type)
{
    var timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
    delay = ((delay !== undefined) ? delay : 0);
    type = ((type !== undefined) ? type : timer.TYPE_ONE_SHOT);
    var callback = {notify: function(timer) { func.apply(obj, args); }};
    timer.initWithCallback(callback, delay, type);
    return timer;
}

function getCallback(obj, func/*, args...*/)
{
    var args = ((arguments.length > 2) ? Array.prototype.slice.call(arguments, 2) : null);
    return function()
    {
        var _args = (args ? Array.prototype.slice.call(arguments).concat(args) : arguments);
        func.apply(obj, _args);
    }
}

function getHandler(name, obj, func)
{
    var handler = {};
    handler[name] = getCallback(obj, func);
    return handler;
}

function initKey(keyId, prefName)
{
    var key = $(keyId);
    for each (var name in ["modifiers", "key", "keycode"])
    {
        var value = nsPreferences.copyUnicharPref(prefName + "." + name);
        if (value)
            key.setAttribute(name, value);
    }
}

function callLater(delay, obj, func/*, args...*/)
{
    if (! func)
        throw new Error("callLater: function is " + func);
    
    var timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
    var args = Array.prototype.slice.call(arguments, 3);
    var callback = {notify: function() { func.apply(obj, args); }}
    timer.initWithCallback(callback, delay, Ci.nsITimer.TYPE_ONE_SHOT);
    return timer;
}


function Timer(precise, delay, obj, func, args)
{
    this.type = (precise ? Ci.nsITimer.TYPE_REPEATING_PRECISE : Ci.nsITimer.TYPE_REPEATING_SLACK);
    this.delay = delay;
    this.callback = {notify: function(timer) { func.apply(obj, args); }};
    this.timer = null;
}

Timer.prototype =
{
    start: function(force)
    {
        if (this.timer)
            if (force)
                this.stop();
            else
                throw new Error("Timer already started");
        this.timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
        this.timer.initWithCallback(this.callback, this.delay, this.type);
    },
    
    stop: function()
    {
        if (this.timer)
        {
            this.timer.cancel();
            this.timer = null;
        }
    },
    
    trigger: function(noReInit)
    {
        if (! noReInit)
            this.stop();
        this.callback.notify(null);
        if (! noReInit)
            this.start();
    }
}


function Box(/*box; width, height*/)
{
    var n = arguments.length;
    var width  = ((n == 1) ? arguments[0].width  : arguments[0]);
    var height = ((n == 1) ? arguments[0].height : arguments[1]);
    
    this.width = width;
    this.height = height;
}

Box.prototype =
{
    contains: function(box, partially)
    {
        if (partially)
            return ((box.width <= this.width) || (box.height <= this.height));
        else
            return ((box.width <= this.width) && (box.height <= this.height));
    },
    
    enlarge: function(box, partially)
    {
        if (! this.contains(box, partially))
        {
            this.width  = (partially ? ((box.width > box.height) ? this.width : box.width)   : Math.max(this.width,  box.width));
            this.height = (partially ? ((box.width > box.height) ? box.height : this.height) : Math.max(this.height, box.height));
        }
        
        return this;
    },
    
    shrink: function(box, partially)
    {
        if (! box.contains(this, partially))
        {
            this.width  = (partially ? ((box.width > box.height) ? box.width : this.width)   : Math.min(this.width,  box.width));
            this.height = (partially ? ((box.width > box.height) ? this.height : box.height) : Math.min(this.height, box.height));
        }
        
        return this;
    },
    
    toString: function()
    {
        return "[" + this.width + ", " + this.height + "]";
    }
}


// nsITreeView
function TreeView(model)
{
    this.model = model;
}

TreeView.prototype =
{
    get rowCount() { return this.model.length; },
    setTree: function(treeBoxObject) { this.treeBoxObject = treeBoxObject; },
    getCellText: function(row, col) { return this.model[row]; },
    isContainer: function(row) { return false; },
    isSeparator: function(row) { return false; },
    isSorted: function() { return false; },
    getLevel: function(row) { return 0; },
    getImageSrc: function(row, col) { return null; },
    getRowProperties: function(row, props) {},
    getCellProperties: function(row, col, props) {},
    getColumnProperties: function(colid, col, props) {}
}


function FilteredTreeView(model, filter) // extend TreeView
{
    this.original = model;
    this._filter = filter;
    this.index2row = new Array(this.original.length);
    
    TreeView.call(this, this._updateModel());
}

FilteredTreeView.prototype =
{
    get originalCount() { return this.original.length; },
    
    get filter() { return this._filter; },
    set filter(value)
    {
        if (value != this._filter)
        {
            this._filter = value;
            this._updateView();
        }
        
        return value;
    },
    
    isIndexFiltered: function(index)
    {
        return (this.index2row[index] === null);
    },
    
    getIndexRow: function(index)
    {
        return this.index2row[index];
    },
    
    _updateModel: function()
    {
        var filter = this.filter;
        var original = this.original;
        var index2row = this.index2row;
        var model = null;
        
        if (filter)
        {
            model = [];
            var row = 0;
            for (var index = 0; index < original.length; index++)
            {
                var element = original[index];
                var shown = (! filter(element));
                if (shown)
                    model.push(element);
                index2row[index] = (shown ? row++ : null);
            }
        }
        else
        {
            model = original;
            for (var index = 0; index < original.length; index++)
                index2row[index] = index;
        }
        
        return this.model = model;
    },
    
    _updateView: function()
    {
        // original, source (old model), destination (new model)
        var org = this.original;
        var src = this.model;
        var dst = this.model = this._updateModel();
        
        var org_n = org.length;
        var src_n = src.length;
        var dst_n = dst.length;
        
        if ((dst_n == src_n) && ((dst_n == 0) || (dst_n == org_n) || arrayEquals(dst, src)))
            return;
        
        var boxObject = this.treeBoxObject;
        var selection = this.selection;
        
        var selectEventsSuppressed = selection.selectEventsSuppressed;
        selection.selectEventsSuppressed = true;
        boxObject.beginUpdateBatch();
        
        var org_i = org_n - 1;
        var src_i = src_n - 1;
        var dst_i = dst_n - 1;
        
        while ((src_i >= 0) && (dst_i >= 0))
        {
            var org_e = org[org_i];
            var src_e = src[src_i];
            var dst_e = dst[dst_i];
            
            var in_src = (src_e === org_e);
            var in_dst = (dst_e === org_e);
            
            if (in_src != in_dst)
                if (in_src)
                    boxObject.rowCountChanged(src_i, -1);
                else
                    boxObject.rowCountChanged(src_i + 1, 1);
            
            org_i--;
            if (in_src) src_i--;
            if (in_dst) dst_i--;
        }
        
        if (src_i >= 0)
            boxObject.rowCountChanged(0, -(src_i + 1));
        if (dst_i >= 0)
            boxObject.rowCountChanged(0, (dst_i + 1));
        
        boxObject.endUpdateBatch();
        selection.selectEventsSuppressed = selectEventsSuppressed;
    }
}

extend(FilteredTreeView, TreeView);
